from typing import Optional, Iterable
import logging
from pandas import DataFrame
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatch
import seaborn as sns
from blacksheep import catheat
from blacksheep._constants import *


def _get_sample_order(annotations: DataFrame, col_of_interest: str) -> Iterable:
    """Orders the samples using annotations

    Args:
        annotations: comparisons vs samples annotation DataFrame
        col_of_interest: Comparison to sort by first

    Returns:
        Sorted order of samples

    """

    sort_by = [col for col in annotations.index if col != col_of_interest]
    annotations = annotations.sort_values([col_of_interest] + sort_by, axis=1)
    return annotations.columns


def _get_genes(qvals: DataFrame, fdr: float, col: str) -> list:
    """Collects significant genes

    Args:
        qvals: qvalues DataFrame
        fdr: FDR cut off
        col: Column for which to collect genes

    Returns:
        List of significant genes

    """

    return list(qvals.loc[(qvals[col] < fdr), :].index)


def _pick_color(red_or_blue: str):
    """Sets colormap for heatmap

    Args:
        red_or_blue: Use red or blue.

    Returns:
        Colormap

    """
    if red_or_blue == "red":
        cmap = sns.cubehelix_palette(
            start=0.857,
            rot=0.00,
            gamma=1.5,
            hue=1,
            light=1,
            dark=0.2,
            reverse=False,
            as_cmap=True,
        )
    elif red_or_blue == "blue":
        cmap = sns.cubehelix_palette(
            start=3,
            rot=0.00,
            gamma=1.5,
            hue=1,
            light=1,
            dark=0.2,
            reverse=False,
            as_cmap=True,
        )
    else:
        raise ValueError("Invalid color choice, must be red or blue, setting to red.")

    cmap.set_bad("#BDBDBD")
    return cmap


def _check_colors(colors: dict) -> dict:
    """Makes sure every input color can be used as a color by the heatmap and legend.

    Args:
        colors: Dictionary of {values: colors}

    Returns:
        Dictionary of values: colors with invalid ones removed.

    """
    for lab, color in colors.items():
        try:
            mpatch.Patch(color=color)
        except ValueError:
            logging.warning("%s is not a valid color" % color)
            colors.pop(lab)
    return colors


def _assign_colors(data: DataFrame, cmap: dict, palette) -> dict:
    """Combines provided colors, and adds more if needed

    Args:
        data: Annotations heatmap to color
        cmap: Provided colormap
        palette: Palette to use to generate unspecific colors

    Returns:
        Color dictionary for every unique value in annotations.

    """

    unique_values = sorted(np.unique(data.values.astype(str)))
    cmap = {v: c for v, c in cmap.items() if v in unique_values}
    missing_entries = [v for v in unique_values if v not in cmap.keys()]
    colors = catheat._gen_colors(palette, len(missing_entries))
    cmap.update({v: colors[i] for i, v in enumerate(missing_entries)})

    return cmap


def _determine_colors(path: str, annotations: DataFrame) -> dict:
    """Takes a file with value, color pairs and fills out any other needed colors.

    Args:
        path: File path to value, color pairs
        annotations: Annotation DataFrame

    Returns:
        Color dictionary

    """

    if not path:
        return _assign_colors(annotations, {}, default_palette)
    try:
        with open(path, "r") as fh:
            colors = {line.split()[0]: line.split()[1] for line in fh.readlines()}
        colors = _check_colors(colors)
    except FileNotFoundError:
        logging.warning("%s is not a valid file, generating colors" % path)
        colors = {}

    return _assign_colors(annotations, colors, default_palette)


def plot_heatmap(
    annotations: DataFrame,
    qvals: DataFrame,
    col_of_interest: str,
    vis_table: DataFrame,
    fdr: float = 0.05,
    red_or_blue: str = "red",
    output_prefix: str = "outliers",
    colors: Optional[str] = None,
    savefig: bool = False,
) -> list:
    """Plots a heatmap of significantly enriched values for a given comparison.

    Args:
        annotations: Annotations DataFrame, samples as rows, annotations as columns
        qvals: qvalues DataFrame with genes/sites as rows and comparisons as columns
        col_of_interest: Which column from qvalues should be used to find signficant genes
        vis_table: Table to be visualized in heatmap. Index values should correspond to the \
        annotation df index, column names should correspond to qvals df index
        fdr: FDR threshold to for significance
        red_or_blue: Whether heatmap should be in red or blue color scale
        output_prefix:  If saving files, output prefix
        colors: File to find color map for annotation header
        savefig: Whether to save the plot to a pdf

    Returns: [annot_ax, vals_ax, cbar_ax, leg_ax]
        List of matplotlib axs, can be further customized before saving. In order the axes \
        contain: annotation header, the heatmap, the color bar, and the legend.

    """

    annotations = annotations.transpose()

    # Get orders
    annot_label = [col for col in annotations.index if col in col_of_interest][0]
    sample_order = _get_sample_order(annotations, annot_label)
    genes = _get_genes(qvals, fdr, col_of_interest)
    if not genes:
        logging.warning("No significant genes at FDR %s in %s" % (fdr, col_of_interest))
        return None

    annotations = annotations.reindex(sample_order, axis=1)
    vis_table = vis_table.reindex(genes).reindex(sample_order, axis=1)

    # Get colors
    cmap = _pick_color(red_or_blue)
    colors = _determine_colors(colors, annotations)

    # Get label
    label = col_of_interest[10:]

    # Set up figure
    sns.set(font="arial", style="white", color_codes=True, font_scale=1)
    plot_height = min(max((0.19 * (len(annotations) + len(genes))), 2), 11)
    plot_width = min(max((0.15 * len(annotations.columns)), 4), 8.5)
    fig = plt.figure(figsize=(plot_width, plot_height))
    gs = plt.GridSpec(
        figure=fig,
        nrows=3,
        ncols=2,
        width_ratios=[len(annotations.columns), 2],
        height_ratios=[len(annotations)] + [len(vis_table) / 2 for i in range(0, 2)],
        wspace=0.01,
        hspace=0.01,
    )

    annot_ax = plt.subplot(gs[0, 0])
    vals_ax = plt.subplot(gs[1:, 0])
    cbar_ax = plt.subplot(gs[-1, 1])
    leg_ax = plt.subplot(gs[:-1, 1])
    leg_ax.axis("off")

    # Header
    catheat.heatmap(
        annotations,
        cmap=colors,
        ax=annot_ax,
        leg_ax=leg_ax,
        leg_kws=dict(loc=(0, 0.05), facecolor="white", edgecolor="white"),
        xticklabels=False,
        yticklabels=annotations.index,
    )

    annot_ax.set_title(plot_title % col_of_interest, fontsize=14)
    annot_ax.set_yticklabels(annotations.index, rotation=0)
    annot_ax.set_xlabel("")
    annot_ax.set_ylabel("")

    # Values
    sns.heatmap(
        vis_table,
        ax=vals_ax,
        cbar_ax=cbar_ax,
        cmap=cmap,
        vmin=0,
        vmax=1,
        yticklabels=vis_table.index,
        xticklabels=False,
        cbar_kws=dict(label=cbar_label),
    )
    vals_ax.set_yticklabels(vis_table.index, rotation=0)
    vals_ax.set_xlabel("")
    vals_ax.set_ylabel("")

    if savefig:
        fig_path = figure_file_name % (output_prefix, label, fdr)
        logging.info("Saving figure to %s" % fig_path)
        plt.savefig(fig_path, dpi=200, bbox_inches="tight")

    return [annot_ax, vals_ax, cbar_ax, leg_ax]



from typing import List, Optional, Tuple
import logging
import os.path
import pandas as pd
from pandas import DataFrame
from blacksheep.classes import OutlierTable, qValues
from blacksheep._outlierTable import _convert_to_outliers
from blacksheep._outlierTable import _convert_to_counts
from blacksheep.comparisons import _compare_groups
from blacksheep.comparisons import get_sample_lists
from blacksheep._constants import *


def make_outliers_table(
    df: DataFrame,
    iqrs: float = 1.5,
    up_or_down: str = "up",
    aggregate: bool = True,
    save_outlier_table: bool = False,
    save_frac_table: bool = False,
    output_prefix: str = "outliers",
    ind_sep: str = "-",
) -> OutlierTable:
    """Converts a DataFrame of values into an OutlierTable object, which includes a DataFrame
    of outlier and non-outlier count values.

    Args:
        df: Input DataFrame with samples as columns and sites/genes as columns.
        iqrs: The number of inter-quartile ranges (IQRs) above or below the median to consider a \
        value as an outlier.
        up_or_down: Whether to call up or down outliers. Up is above the median; down \
        is below the median. Options "up" or "down".
        aggregate: Whether to sum outliers across a grouping (e.g. gene-level) than individual \
        sites. For instance if columns indicate phosphosites on proteins, with the format \
        "RAG2-S365", output will show counts of outliers per protein (e.g. RAG2) rather than on \
        individual sites (e.g. RAG2-S365).
        save_outlier_table: Whether to write a file with the outlier count table.
        save_frac_table: Whether to write a file with the outlier fraction table.
        output_prefix: If files are written, a prefix for the files.
        ind_sep: The separator used in sites, for instance, to separate a gene and site. \
        If just using genes (i.e. no separator), or not aggregating this parameter has no effect.

    Returns: outliers
        Returns an OutlierTable object, with outlier and non-outlier counts and metadata
        about how the outliers were called.

    """

    samples = df.columns
    logging.info("Calling outliers for %s samples" % len(samples))

    df = _convert_to_outliers(df, samples, iqrs, up_or_down)
    df = _convert_to_counts(df, samples, aggregate, ind_sep)
    outliers = OutlierTable(df, up_or_down, iqrs, samples, None)

    if save_frac_table:
        frac_path = os.path.abspath(frac_table_file_name % (output_prefix, up_or_down))
        logging.info("Saving outlier fraction table to %s" % frac_path)
        outliers.frac_table.to_csv(frac_path, sep="\t")

    if save_outlier_table:
        out_path = os.path.abspath(
            outlier_table_file_name % (output_prefix, up_or_down)
        )
        logging.info("Saving outlier table to %s" % out_path)
        df.to_csv(out_path, sep="\t")

    return outliers


def compare_groups_outliers(
    outliers: OutlierTable,
    annotations: DataFrame,
    frac_filter: Optional[float] = 0.3,
    save_qvalues: bool = False,
    output_prefix: str = "outliers",
    save_comparison_summaries: bool = False,
) -> qValues:
    """Takes an OutlierTable object and a sample annotation DataFrame and performs comparisons for
    any column in annotations with exactly 2 groups. For each group identified in the annotations
    DataFrame, this function will calculate the q-values of enrichment of outliers for each row in
    each group.

    Args:
        outliers: An OutlierTable, with a DataFrame of outlier and non-outlier counts, \
        as well as parameters for how outliers were calculated.
        annotations: A DataFrame with samples as rows and annotations as columns. Each \
        column must contain exactly 2 different categories, not counting missing values. Columns \
        without 2 options will be ignored.
        frac_filter: The fraction of samples in the group of interest that must \
        have an outlier value to be considered in the comparison. Float between 0 and 1 or None.
        save_qvalues: Whether to write a file with a table of qvalues.
        output_prefix: If files are written, a prefix for the files.
        save_comparison_summaries: Whether to write a file for each annotation column with the \
        counts in the fisher table, pvalues and q values per row.

    Returns: qvals
        A qValues object, which includes a DataFrame of q-values for each comparison, \
        as well as some metadata about how the comparisons were performed.


    """

    df = outliers.df
    samples = outliers.samples
    up_or_down = outliers.up_or_down
    results_df = pd.DataFrame(index=df.index)
    for comp in annotations.columns:
        logging.info("Testing for enrichment in %s comparison" % comp)

        group0_label, group0, group1_label, group1 = get_sample_lists(annotations, comp)
        # Checking everything is in place
        if group0 is None:
            logging.error(
                "There are not exactly 2 groups of samples, skipping %s" % comp
            )
            continue
        not_there = [samp for samp in group0 if samp not in samples] + [
            samp for samp in group1 if samp not in samples
        ]
        if not_there:
            logging.warning(
                "These samples were not found in outliers table: "
                "%s, continuing without them. " % ", ".join(not_there)
            )
        group0 = [samp for samp in group0 if samp in samples]
        group1 = [samp for samp in group1 if samp in samples]
        if len(group0) < 2:
            logging.error(
                "Group %s does not have at least two samples, "
                "skipping comparison %s. " % (group0_label, comp)
            )
            continue
        if len(group1) < 2:
            logging.error(
                "Group %s does not have at least two samples, "
                "skipping comparison%s. " % (group1_label, comp)
            )
            continue

        # doing tests
        label0 = fdr_col_label % (comp, group0_label)
        results_df, fisher_info0 = _compare_groups(
            results_df, df, group0, group1, frac_filter, label0
        )

        label1 = fdr_col_label % (comp, group1_label)
        results_df, fisher_info1 = _compare_groups(
            results_df, df, group1, group0, frac_filter, label1
        )

        if save_comparison_summaries:
            fisher_info0.columns = [
                "%s_%s_%s" % (outlier_count_lab, comp, group0_label),
                "%s_%s_%s" % (outlier_count_lab, comp, group1_label),
                "%s_%s_%s" % (not_outlier_count_lab, comp, group0_label),
                "%s_%s_%s" % (not_outlier_count_lab, comp, group1_label),
                specific_fisher_p % (comp, group0_label),
            ]

            fisher_info1.columns = [
                "%s_%s_%s" % (outlier_count_lab, comp, group1_label),
                "%s_%s_%s" % (outlier_count_lab, comp, group0_label),
                "%s_%s_%s" % (not_outlier_count_lab, comp, group1_label),
                "%s_%s_%s" % (not_outlier_count_lab, comp, group0_label),
                specific_fisher_p % (comp, group1_label),
            ]
            comp_df = pd.concat(
                [fisher_info0, fisher_info1], axis=0, join="outer", sort=True
            ).merge(
                results_df.reindex([label0, label1], axis=1),
                left_index=True,
                right_index=True,
            )
            if len(comp_df) > 0:
                comp_df.to_csv(
                    ind_comparison_file_name % (output_prefix, up_or_down, comp),
                    sep="\t",
                )
    results_df = results_df.dropna(how="all", axis=0)
    if save_qvalues:
        qval_path = os.path.abspath(qvalues_file_name % (output_prefix, up_or_down))
        logging.info("Saving qvalues to %s" % qval_path)
        results_df.to_csv(qval_path, sep="\t")
    qvals = qValues(results_df, annotations.columns, frac_filter)
    return qvals


def deva(
    df: DataFrame,
    annotations: DataFrame,
    iqrs: float = 1.5,
    up_or_down: str = "up",
    aggregate: bool = True,
    save_outlier_table: bool = False,
    save_frac_table: bool = False,
    frac_filter: Optional[float] = 0.3,
    save_qvalues: bool = False,
    output_prefix: str = "outliers",
    ind_sep: str = "-",
    save_comparison_summaries: bool = False,
) -> Tuple[OutlierTable, qValues]:
    """
    Takes a DataFrame of values and returns OutlierTable and qValues objects. This command runs
    the whole outliers pipeline. The DataFrame in the OutlierTable object can be used to run more
    comparisons in future. The qValues object can be used for visualization, or writing
    significant gene lists.

    Args:
        df: Input DataFrame with samples as columns and sites/genes as rows.
        annotations: A DataFrame with samples as rows and annotations as columns. Each \
        column must contain exactly 2 different values, not counting missing \
        values. Other columns will be ignored.
        iqrs: The number of interquartile ranges (IQRs) above or below the median to consider a \
        value as an outlier.
        up_or_down: Whether to call up or down outliers. Up is above the median; down \
        is below the median. Options "up" or "down".
        aggregate: Whether to sum outliers across a grouping (e.g. gene-level) than individual \
        sites. For instance if columns indicate phosphosites on proteins, with the format \
        "RAG2-S365", output will show counts of outliers per protein (e.g. RAG2) rather than on \
        individual sites (e.g. RAG2-S365).
        save_outlier_table: Whether to write a file with the outlier count table.
        save_frac_table: Whether to write a file of the fraction of outliers.
        frac_filter: The fraction of samples in the group of interest that must \
        have an outlier value to be considered in the comparison. Float between 0 and 1 or None.
        save_qvalues: Whether to write a file of qvalues.
        output_prefix: If files are written, a prefix for the files.
        ind_sep: The separator used in the columns, for instance, to separate a gene and site. \
        If just using genes (i.e. no separator), or not aggregating this parameter \
        has no effect.
        save_comparison_summaries: Whether to write a table for each comparison with the \
        counts in the fisher table, pvalues and qvalues per row.

    Returns: outliers, qvals
        Returns an OutlierTable object and qValues object.

    """

    logging.info("Making outliers table")
    outliers = make_outliers_table(
        df,
        iqrs,
        up_or_down,
        aggregate,
        save_outlier_table,
        save_frac_table,
        output_prefix,
        ind_sep,
    )

    logging.info("Performing group comparisons")
    qvals = compare_groups_outliers(
        outliers,
        annotations,
        frac_filter,
        save_qvalues,
        output_prefix,
        save_comparison_summaries,
    )

    return outliers, qvals



#    This script is part of pymaid (http://www.github.com/schlegelp/catheat).
#    Copyright (C) 2017 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along


from typing import Optional
import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.axes
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches


def heatmap(
    data,
    cmap: Optional[dict] = None,
    palette: str = "hls",
    ax=None,
    legend: bool = True,
    leg_pos: str = "right",
    leg_ax=None,
    leg_kws: Optional[dict] = None,
    **sns_kws
):
    """ Class to plot categorical heatmap using seaborn.

    Parameters
    ----------
    data :      rectangular dataset
                2D dataset that can be coerced into an ndarray. If a Pandas
                DataFrame is provided, the index/column information will be
                used to label the columns and rows.
    cmap :      dict, optional
                Colors for each category in the dataset. Missing colors will
                be added from the palette.
    palette :   matplotlib/seaborn color palette name or object, optional
                Palette to be used for heatmap.
    ax :        matplotlib ax, optional
    legend :    bool, optional
                If True, plot legend.
    leg_ax :    matplotlib axis, optional
                By default, will add legend to same ax as heatmap. Use this
                argument to explicitly set legend ax.
    leg_pos :   {'right', 'top'}
                Position of legend. Only relevant if legend ax is not
                explicitly provided via `leg_ax`.
    leg_kws :   dict, optional
                Keyword arguments passed to plt.legend()

    **sns_kws
                Keyword argumentas passed through to `seaborn.heatmap()`

    Returns
    -------
    matplotlib axis
                    Heatmap axis as returned by seaborn.heatmap().
    colormap :      dict
                    Colormap mapping categorical values to RGB colours.

    """
    if cmap is None:
        cmap = {}
    if leg_kws is None:
        leg_kws = {}
    if not isinstance(data, (pd.DataFrame, pd.Series, np.ndarray)):
        raise TypeError('Unable to work with data of type "{0}"'.format(type(data)))

    # If not provided, get ax
    if ax is None:
        _, ax = plt.subplots()

    # Get unique values in dataset
    if isinstance(data, (pd.DataFrame, pd.Series)):
        unique_values = sorted(np.unique(data.values.astype(str)))
    else:
        unique_values = sorted(np.unique(data.astype(str)))

    n_unique = len(unique_values)

    # Prepare colors
    if not cmap:
        # Generate colors -> for the heatmap
        colors = _gen_colors(palette, n_unique)

        # We have colours, let's generate a cmap -> for the legend
        cmap = {v: colors[i] for i, v in enumerate(unique_values)}
    elif isinstance(cmap, dict):
        # Check for missing entries
        missing_entries = [v for v in unique_values if v not in cmap]

        # Generate missing colors
        colors = _gen_colors(palette, len(missing_entries))

        # Update colormap
        cmap.update({v: colors[i] for i, v in enumerate(missing_entries)})

        # Generate colours in order -> for legend
        colors = [cmap[cat] for cat in unique_values]

    else:
        raise TypeError('Unable to intepret colormap of type "{0}"'.format(type(cmap)))

    # Turn data into numeric values
    vmap = {c: i for i, c in enumerate(unique_values)}

    mapper = lambda t: vmap[str(t)]
    vfunc = np.vectorize(mapper)

    if isinstance(data, pd.DataFrame):
        numerical_data = data.applymap(mapper)
    elif isinstance(data, np.ndarray):
        numerical_data = vfunc(data)

    # Plot heatmap
    cmap_object = mcolors.LinearSegmentedColormap.from_list(
        "custom", colors, N=len(colors)
    )
    sns_ax = sns.heatmap(numerical_data, ax=ax, cmap=cmap_object, cbar=False, **sns_kws)

    # Add legend
    if legend:
        if not leg_ax:
            # Use the heatmap's ax if none provided
            leg_ax = sns_ax

            # Make heatmap a bit less wide
            box = sns_ax.get_position()
            sns_ax.set_position([box.x0, box.y0, box.width * 0.9, box.height])

            # Add some default specific to using the heatmap for legend
            if leg_pos.lower() == "top":
                leg_kws["bbox_to_anchor"] = (0.0, 1.02, 1.0, 0.102)
                leg_kws["ncol"] = n_unique
            elif leg_pos.lower() == "right":
                leg_kws["bbox_to_anchor"] = (1.02, 0.0, 0.102, 1.0)
                leg_kws["ncol"] = 1
            leg_kws["fontsize"] = 8
            leg_kws["mode"] = "expand"

        elif not isinstance(leg_ax, matplotlib.axes.Axes):
            raise TypeError(
                'leg_ax must be matplotlib axes, not "{}"'.format(type(leg_ax))
            )

        patches = [
            mpatches.Patch(facecolor=cmap[v], edgecolor="black") for v in unique_values
        ]

        _ = leg_ax.legend(patches, unique_values, **leg_kws)

    return sns_ax, cmap


def _is_categorical(x):
    """Return True if data is categorical (i.e. not numerical)."""
    if isinstance(x, pd.DataFrame):
        num_cols = x.mean(axis=0).index.tolist()
        return [col not in num_cols for col in x.columns]
    if isinstance(x, pd.Series):
        try:
            x.mean()
            return True
        except BaseException:
            return False
    if isinstance(x, np.ndarray):
        if x.ndim > 2:
            raise ValueError("Can only process 1d or 2d arrays.")
        if x.ndim == 2:
            cat_cols = []
            for i in range(x.shape[1]):
                try:
                    x[:, i].astype(int)
                    cat_cols.append(False)
                except BaseException:
                    cat_cols.append(True)
        elif x.ndim == 1:
            try:
                x.astype(int)
                return False
            except BaseException:
                return True


def _gen_colors(pal, n):
    """Generate colours from provided palette."""

    # If string
    if isinstance(pal, str):
        # First, check if seaborn palette
        try:
            colors = sns.color_palette(pal, n)
        # If not, try getting the matplotlib palette
        except BaseException:
            pal = plt.get_cmap(pal)
            colors = [pal(i) for i in np.linspace(0, 1, n)]
    # If palette provided
    elif isinstance(pal, mcolors.LinearSegmentedColormap):
        colors = [pal(i) for i in np.linspace(0, 1, n)]
    # If list of colors
    elif isinstance(pal, (list, np.ndarray)):
        if len(pal) < n:
            raise ValueError(
                "Must provide at least as many colors as there are unique entries: {0}".format(
                    len(n)
                )
            )
        colors = pal
    else:
        raise TypeError(
            'Unable to generate colors from palette of type "{0}"'.format(type(pal))
        )

    return colors



import warnings
import logging
import os.path
from typing import Iterable, Optional
import pandas as pd
from pandas import DataFrame, Series
import numpy as np
import sklearn.linear_model as lm
from sklearn.exceptions import UndefinedMetricWarning
from blacksheep.classes import OutlierTable
from blacksheep._constants import *


def _is_valid_file(arg: str) -> str:
    """Checks if file exists (probably pretty redundant except as a type checker in argparse)

    Args:
        arg: File path

    Returns: arg
        Validated file path

    """
    if not os.path.exists(arg):
        raise FileNotFoundError("%s does not exist" % arg)
    return arg


def _check_output_prefix(arg: str) -> str:
    """Checks if output prefix is valid

    Args:
        arg: Output prefix

    Returns: arg
        Validation output prefix

    """

    if "/" in arg:
        prefix = arg.rsplit("/", 1)[0]
        _is_valid_file(prefix)
    return arg


def _check_suffix(path: str) -> str:
    """Checks that file is a .csv or .tsv file, and returns which sep to use

    Args:
        path: File path

    Returns: sep
        Sep to use for parsing

    """

    if path[-4:] == ".tsv":
        return "\t"
    if path[-4:] == ".csv":
        return ","
    raise ValueError("File must be .csv or .tsv")


def read_in_values(path: str) -> DataFrame:
    """Figures out sep and parsing file into dataframe.

    Args:
        path: File path

    Returns: df
        DataFrame from table in file

    """
    sep = _check_suffix(path)
    return pd.read_csv(_is_valid_file(path), sep=sep, index_col=0)


def read_in_outliers(path: str, updown: str, iqrs: float) -> OutlierTable:
    """Parses a file into an OutlierTable object.

    Args:
        path: File path
        updown: Whether the outliers represent up or down outliers
        iqrs: How many IQRs were used to define an outlier

    Returns: outliers
        OutlierTable object

    """

    sep = _check_suffix(path)
    df = pd.read_csv(_is_valid_file(path), sep=sep, index_col=0)
    samples = sorted(list(set([ind.rsplit(col_seps, 1)[0] for ind in df.columns])))
    return OutlierTable(df, updown, iqrs, samples, None)


def binarize_annotations(df: DataFrame) -> DataFrame:
    """Takes an annotation DataFrame, checks each column for the number of possible values,
    and adjusts based on that. If the column has 0 or 1 options, it is dropped. Cols with 2
    possible values are retained as-is. Cols with more than 2 values are expanded. For each
    value in that column, a new column is created with val and not_val options.

    Args:
        df: Annotations DataFrame.

    Returns: new_df
        Refactored annotations DataFrame.
    """

    new_df = pd.DataFrame(index=df.index)
    for col in df.columns:
        if len(df[col].dropna().value_counts().keys()) == 2:
            new_df[col] = df[col]
        elif len(df[col].dropna().value_counts().keys()) > 2:
            for val in df[col].dropna().value_counts().keys():
                val = str(val).replace("_", "-")
                new_df.loc[(df[col] != val), binarized_col_name % (col, val)] = (
                    outgroup_val % val
                )
                new_df.loc[(df[col] == val), binarized_col_name % (col, val)] = val
                new_df.loc[df[col].isnull(), binarized_col_name % (col, val)] = np.nan
    return new_df


def _convert_to_residuals(valstarget: Series, valsnormer: Series, model) -> Series:
    """ Take a 1d array of values to normalize, and a 1d array of values to use for
    normalization and a model. Trains the model using the target as labels and normer as
    features, then predicts for each normer values and subtracts the original target value,
    leaving residual values.

    Args:
        valstarget: 1d array of values to normalize
        valsnormer: 1d array of values to use to normalize
        model: a machine learning model, on which you can perform .fit and .predict

    Returns: residuals

    """
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    warnings.filterwarnings("ignore", category=UndefinedMetricWarning)
    nonull = (valstarget.isnull() == False) & (valsnormer.isnull() == False)
    if sum(nonull) < model.get_params()["cv"]:
        residuals = np.empty(len(valstarget))
        return pd.Series(residuals, index=valstarget.index)
    features = valsnormer[nonull].values.reshape(-1, 1)
    labels = valstarget[nonull].values
    model = model.fit(features, labels)
    prediction = model.predict(features)
    residuals = labels - prediction
    return pd.Series(residuals, index=valstarget[nonull].index)


def normalize_df(
    target: DataFrame,
    normer: DataFrame,
    ind_sep: Optional[str] = "-",
    alphas: Optional[Iterable[float]] = None,
    cv: float = 5,
    **RidgeCV_kws
) -> DataFrame:
    """ Used to normalize a dataset by another dataset, using a linear model with regularization
    chosen through cross validation (aka sklearn's RidgeCV). This is useful for normalizing,
    for example, RNA values by CNA, or phosphopeptide values by protein abundance. If target and
    normer dataframe row IDs (index) match 1:1, pass None for ind_sep.

    Args:
        target: Dataframe of values to normalize. Row IDs (index) before the sep (or whole ID
        if no sep) must match normer IDs. Row IDs must be unique.
        normer: Dataframe of values to use for normalization. Row IDs must match all or
        pre-ind_sep portions of target row IDs. Row IDs must be unique.
        ind_sep: If multiple rows in target map to 1 row in normer, the delimiter used to split
        the unique ID that matches the normer IDs. Defaul "-"
        alphas: Parameters to try for regulariztion. If None, tries powers of 2 from -10 to 10.
        cv: Fold for cross validation. Also the minimum number of non-null values for each
        row. Default 5
        **RidgeCV_kws: kws to pass to sklearn's RidgeCV

    Returns: normed
        The target dataframe normalized by the normer dataframe. Only includes rows with
        sufficient non-null values from both dataframe.

    """

    if not alphas:
        alphas = [2 ** i for i in range(-10, 10, 1)]

    normer = normer[[col for col in target.columns if col in normer.columns]]
    target = target[normer.columns]
    if (len(normer.columns) < cv) or (len(target.columns) < cv):
        raise KeyError("target and normer dataframes do not have at least %s columns in common" %cv)

    target = target.transpose()
    target["col0"] = 0
    target.set_index("col0", append=True, inplace=True)
    target = target.reorder_levels(
        [target.index.names[-1], target.index.names[0]]
    ).transpose()

    normer = normer.transpose()
    normer["col0"] = 1
    normer.set_index("col0", append=True, inplace=True)
    normer = normer.reorder_levels(
        [normer.index.names[-1], normer.index.names[0]]
    ).transpose()

    target["gene"] = [i.split(ind_sep)[0] for i in target.index]
    target = target.loc[target["gene"].isin(normer.index), :]
    if len(target) == 0:
        raise KeyError("No rows in common between target and normer")
    logging.info(
        "Normalizing %s common rows and %s common samples between target and normer"
        %(len(target), len(normer.columns))
    )
    data = target.merge(normer, how="left", left_on="gene", right_index=True)

    model = lm.RidgeCV(alphas=alphas, cv=cv, **RidgeCV_kws)
    normed = data.apply(
        (lambda row: _convert_to_residuals(row[0], row[1], model)), axis=1
    )

    return normed



import logging
from blacksheep.classes import qValues, OutlierTable
from blacksheep.deva import make_outliers_table, compare_groups_outliers, deva
from blacksheep.visualization import plot_heatmap
from blacksheep.parsers import (
    binarize_annotations,
    normalize_df,
    read_in_values,
    read_in_outliers,
)


FMT = "%(asctime)s:%(levelname)s:%(message)s"
logging.basicConfig(format=FMT, level=logging.WARNING, datefmt="%m/%d/%Y %H:%M:%S")


__all__ = [
    "make_outliers_table",
    "compare_groups_outliers",
    "deva",
    "plot_heatmap",
    "binarize_annotations",
    "normalize_df",
    "read_in_values",
    "read_in_outliers",
    "qValues",
    "OutlierTable"
]



import pandas as pd
import numpy as np
import scipy.stats
from pandas import DataFrame
from typing import List
from blacksheep._constants import *


SampleList = List[str]


def _convert_to_outliers(
    df: DataFrame, samples: SampleList, num_iqrs: float, up_or_down: str
) -> DataFrame:
    """Calls outliers on a given values table.

    Args:
        df: Input DataFrame with samples as columns and genes or sites as rows. \
        Index should be an identifier for each row.
        samples: List of samples to be considered in the distribution when defining median, \
        IQR and outliers.
        num_iqrs: How many inter-quartile ranges (IQRs) above or below the median to consider \
        something an outlier.
        up_or_down: Whether to call outliers above the median (up) or below the median (down)

    Returns:
        A DataFrame with outlier calls for each value. 0 means not an outlier; 1 means there is \
        an outlier. Missing values are propagated.

    """
    if num_iqrs <= 0:
        raise ValueError("num_iqrs must be greater than 0")

    df = df.copy()
    df[row_iqr_name] = scipy.stats.iqr(df[samples], axis=1, nan_policy="omit")
    df[row_median_name] = np.nanquantile(df[samples], q=0.5, axis=1)

    outlier_df = pd.DataFrame()

    if up_or_down == "up":
        df[row_upper_bound_name] = df[row_median_name] + (num_iqrs * df[row_iqr_name])
        outlier_df[samples] = (
            df[samples].gt(df[row_upper_bound_name], axis=0).astype(int)
        )
        outlier_df[df[samples].isnull()] = np.nan
        return outlier_df
    elif up_or_down == "down":
        df[row_lower_bound_name] = df[row_median_name] - (num_iqrs * df[row_iqr_name])
        outlier_df[samples] = (
            df[samples].lt(df[row_lower_bound_name], axis=0).astype(int)
        )
        outlier_df[df[samples].isnull()] = np.nan
        return outlier_df
    raise ValueError("up_or_down must be either 'up' or 'down'")


def _convert_to_counts(
    df: DataFrame, samples: SampleList, aggregate: bool, ind_sep: str
) -> DataFrame:
    """Counts outliers and non-outlier values for each sample and each row (if aggregate=False)
    or each unique identifier (if aggregate=True).

    Args:
        df: Outlier DataFrame from convertToOutliers function.
        samples: List of samples to consider. Should be same list as input to convertToOutliers.
        aggregate: Whether or not to collapse multiple rows with identical identifiers before \
        a separater. Collapsing values is done by counting outliers and non-outliers for each \
        sample per unique identifier.
        ind_sep: The separator used in the index to separate a more general ID and less specific \
        ID. e.g. in RAG2-S365 the separater is "-".

    Returns:
        Outlier DataFrame that has a column with counts of non-outliers and outliers for \
        each sample, with a row for each input site (if no aggregation) or each unique identifier \
        (with aggregation). This table is the input for the comparison function.

    """
    not_outlier_cols = [x + col_seps + col_not_outlier_suffix for x in samples]
    outlier_cols = [x + col_seps + col_outlier_suffix for x in samples]

    if aggregate:
        df.index = [ind.split(ind_sep)[0] for ind in df.index]
        df_inv = df == 0

        output_df = pd.DataFrame()
        output_df[not_outlier_cols] = df_inv.groupby(level=0)[samples].sum()
        output_df[outlier_cols] = df.groupby(level=0)[samples].sum()
    elif not aggregate:
        output_df = pd.DataFrame(index=df.index)
        output_df[outlier_cols] = df[samples]
        output_df[not_outlier_cols] = 1 - df[samples]
    return output_df



from typing import Optional, List, Iterable
import sys
import logging
import argparse

import matplotlib.pyplot as plt
from blacksheep.deva import deva
from blacksheep.deva import make_outliers_table
from blacksheep.deva import compare_groups_outliers
from blacksheep import parsers
from blacksheep.parsers import _is_valid_file, _check_output_prefix
from blacksheep.classes import qValues
from blacksheep.visualization import plot_heatmap
from blacksheep._constants import *


fmt = "%(asctime)s:%(levelname)s:%(message)s"
logging.basicConfig(format=fmt, level=logging.INFO, datefmt="%m/%d/%Y %H:%M:%S")


def _set_up_logger(path):
    logger = logging.getLogger("cli")
    fh = logging.FileHandler("%s.log" % path)
    fh.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARNING)
    fmter = logging.Formatter(fmt)
    fh.setFormatter(fmter)
    ch.setFormatter(fmter)
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger


# arg checkers
def _check_positive(arg: str) -> float:
    try:
        arg = float(arg)
    except TypeError:
        raise argparse.ArgumentTypeError("%s is not a valid number" % arg)
    if arg < 0:
        raise argparse.ArgumentTypeError("%s is not a positive number" % arg)
    return arg


def _bn0and1(arg: str) -> float:
    try:
        arg = float(arg)
    except TypeError:
        raise argparse.ArgumentTypeError("%s is an invalid positive value" % arg)
    if (arg < 0) or (arg > 1):
        raise argparse.ArgumentTypeError("%s is not a number between 0 and 1" % arg)
    return arg


# Argparser
def _make_parser():
    parser = argparse.ArgumentParser(prog="blacksheep", description="")
    parser.add_argument("--version", "-v", action="version", version="%(prog)s 0.0.1")

    subparsers = parser.add_subparsers(dest="which")
    subparsers.required = True

    outliers_table = subparsers.add_parser(
        "outliers_table",
        description="Takes a table of values and converts to a "
        "table of outlier counts.",
    )
    outliers_table.add_argument(
        "values",
        type=_is_valid_file,
        help="File path to input values. Columns must be samples, "
        "genes must be sites or genes. Only .tsv and .csv accepted.",
    )
    outliers_table.add_argument(
        "--output_prefix",
        type=_check_output_prefix,
        default="outliers",
        help="Output prefix for writing files. Default outliers. ",
    )
    outliers_table.add_argument(
        "--iqrs",
        type=_check_positive,
        default=1.5,
        help="Number of interquartile ranges (IQRs) above or below the "
        "median to consider a value an outlier. Default is 1.5 IQRs.",
    )
    outliers_table.add_argument(
        "--up_or_down",
        type=str,
        default="true",
        choices=["up", "down"],
        help="Whether to look for up or down outliers. Choices are up or "
        "down. Default up.",
    )
    outliers_table.add_argument(
        "--ind_sep",
        type=str,
        default="-",
        help="If site labels have a parent molecule (e.g. a gene name such "
        "as ATM) and a site identifier (e.g. S365) this is the "
        "delimiter between the two elements. Default is -",
    )
    outliers_table.add_argument(
        "--do_not_aggregate",
        default=False,
        action="store_true",
        help="Use flag if you do not want to sum outliers based on site " "prefixes.",
    )
    outliers_table.add_argument(
        "--write_frac_table",
        default=False,
        action="store_true",
        help="Use flag if you want to write a table with fraction of "
        "values per site, per sample that are outliers. Will not be "
        "written by default. Useful for visualization. ",
    )

    binarize = subparsers.add_parser(
        "binarize",
        description="Takes an annotation table where some columns "
        "may have more than 2 possible values (not "
        "including empty/null values) and outputs an "
        "annotation table with only two values per "
        "annotation. Propagates null values. ",
    )
    binarize.add_argument(
        "annotations",
        type=_is_valid_file,
        help="Annotation table with samples as rows and "
        "annotation labels as columns. ",
    )
    binarize.add_argument(
        "--output_prefix",
        type=_check_output_prefix,
        default="annotations",
        help="Output prefix for writing files. Default annotations. Suffix will be '.binarized.tsv'",
    )

    normalize = subparsers.add_parser(
        "normalize",
        description="Takes a target table and a normalizer table, and returns a normalized "
                    "target table. Builds a regularized linear model for each line in the target "
                    "table using the matching row ID in the normalizer table, and finds the "
                    "residuals of that model for each value. for example, this could be used to "
                    "normalize phospho-peptide data by protein abundance data; resulting values "
                    "will reflect only abundance differences due to phosphorylation changes, "
                    "not peptide abundances. Another use could be normalizing RNA by CNA. "
    )
    normalize.add_argument(
        "target_values",
        type=_is_valid_file,
        help="Table of values to be normalized. Sites/genes as rows, samples as columns. "
             "Row identifiers must be unique. "
    )
    normalize.add_argument(
        "normalizer_values",
        type=_is_valid_file,
        help="Table of values to use for normalization. Sites/genes as rows, samples as columns. "
             "Row identifiers must be unique, and must match the pre-ind_sep part of the target "
             "values identifiers."
    )
    normalize.add_argument(
        "--ind_sep",
        type=str,
        default=None,
        help="Separator used in index if target is site specific. Row IDs before ind_sep in "
             "the target must match the row IDs in normalizer_values. If row IDs already match, "
             "leave blank."
    )
    normalize.add_argument(
        "--output_prefix",
        type=str,
        default="values",
        help="Prefix for output file. Suffix will be '.normalized.tsv'"
    )

    compare_groups = subparsers.add_parser(
        "compare_groups",
        description="Takes an annotation table and outlier count "
        "table (output of outliers_table) and outputs qvalues from "
        "a statistical test that looks for enrichment of outlier values in each "
        "group in the annotation table. For each value in each comparison, the qvalue table will "
        "have 1 column, if there are any genes in that comparison. "
    )
    compare_groups.add_argument(
        "outliers_table",
        type=_is_valid_file,
        help="Table of outlier counts (output of outliers_table). Must be "
        ".tsv or .csv file, with outlier and non-outlier counts as columns, and genes/sites as "
        "rows. ",
    )
    compare_groups.add_argument(
        "annotations",
        type=_is_valid_file,
        help="Table of annotations. Must be .csv or .tsv. Samples as rows "
        "and comparisons as columns. Comparisons must have only  "
        "unique values (not including missing values). If there are "
        "more options than that, you can use binarize to prepare the "
        "table. ",
    )
    compare_groups.add_argument(
        "--output_prefix",
        type=_check_output_prefix,
        default="outliers",
        help="Output prefix for writing files. Default outliers. ",
    )
    compare_groups.add_argument(
        "--frac_filter",
        type=_bn0and1,
        default=0.3,
        help="The minimum fraction of samples per group that must have an outlier in a gene to"
        "consider that gene in the analysis. This is used to prevent a high number of outlier "
        "values in 1 sample from driving a low qvalue. Default 0.3",
    )
    compare_groups.add_argument(
        "--write_comparison_summaries",
        default=False,
        action="store_true",
        help="Use flag to write a separate file for each column in the annotations table, "
        "with outlier counts in each group, p-values and q-values in each group. ",
    )
    compare_groups.add_argument(
        "--iqrs",
        type=_check_positive,
        default=None,
        help="Number of IQRs used to define outliers in the input count table. Optional.",
    )
    compare_groups.add_argument(
        "--up_or_down",
        type=str,
        choices=["up", "down"],
        help="Whether input outlier table represents up or down outliers. Needed for "
        "output file labels. Default up",
    )
    compare_groups.add_argument(
        "--write_gene_list",
        default=False,
        action="store_true",
        help="Use flag to write a list of significantly enriched genes for each value in each "
        "comparison. If used, need an fdr threshold as well. ",
    )
    compare_groups.add_argument(
        "--make_heatmaps",
        default=False,
        action="store_true",
        help="Use flag to draw a heatmap of signficantly enriched genes for each value in each "
        "comparison. If used, need an fdr threshold as well. ",
    )
    compare_groups.add_argument(
        "--fdr",
        type=_bn0and1,
        default=0.05,
        help="FDR cut off to use for signficantly enriched gene lists and heatmaps. Default 0.05",
    )
    compare_groups.add_argument(
        "--red_or_blue",
        type=str,
        choices=["red", "blue"],
        default="red",
        help="If --make_heatmaps is called, color of values to draw on heatmap. Default red. ",
    )
    compare_groups.add_argument(
        "--annotation_colors",
        type=_is_valid_file,
        default=None,
        help="File with color map to use for annotation header if --make_heatmaps is used. Must "
        "have a 'value    color' format for each value in annotations. Any value not "
        "represented will be assigned a new color. ",
    )

    visualize = subparsers.add_parser(
        "visualize",
        description="Used to make custom heatmaps from significant " "genes. ",
    )
    visualize.add_argument(
        "comparison_qvalues",
        type=_is_valid_file,
        help="Table of qvalues, output from compare_groups. Must be .csv or .tsv. Has genes/sites "
        "as rows and comparison values as columns. ",
    )
    visualize.add_argument(
        "annotations",
        type=_is_valid_file,
        help="Table of annotations used to generate qvalues. ",
    )
    visualize.add_argument(
        "visualization_table",
        type=_is_valid_file,
        help="Values to visualize in heatmap. Samples as columns and "
        "genes/sites as rows. Using outlier fraction table is recommended, but original "
        "values can also be used if no aggregation was used. ",
    )
    visualize.add_argument(
        "comparison_of_interest",
        type=str,
        help="Name of column in qvalues table from which to visualize significant genes. ",
    )
    visualize.add_argument(
        "--output_prefix",
        type=_check_output_prefix,
        default="outliers",
        help="Output prefix for writing files. Default outliers. ",
    )
    visualize.add_argument(
        "--annotations_to_show",
        type=str,
        default=None,
        nargs="+",
        help="Names of columns from the annotation table to show in the header of the heatmap. "
        "Default is all columns. ",
    )
    visualize.add_argument(
        "--fdr",
        type=_bn0and1,
        default=0.05,
        help="FDR threshold to use to select genes to visualize. Default 0.05",
    )
    visualize.add_argument(
        "--red_or_blue",
        type=str,
        choices=["red", "blue"],
        default="red",
        help="Color of values to draw on heatmap. Default red. ",
    )
    visualize.add_argument(
        "--annotation_colors",
        type=_is_valid_file,
        default=None,
        help="File with color map to use for annotation header. Must "
        "have a line with 'value    color' format for each value in annotations. Any value "
        "not represented will be assigned a new color. ",
    )
    visualize.add_argument(
        "--write_gene_list",
        default=False,
        action="store_true",
        help="Use flag to write a list of significantly enriched genes for each value in each "
        "comparison.",
    )

    outliers = subparsers.add_parser(
        "deva",
        description="Runs whole outliers pipeline. Has options to output every "
        "possible output. ",
    )
    outliers.add_argument(
        "values",
        type=_is_valid_file,
        help="File path to input values. Samples are columns and genes/sites are rows. Only .tsv "
        "and .csv accepted.",
    )
    outliers.add_argument(
        "annotations",
        type=_is_valid_file,
        help="File path to annotation values. Rows are sample names, "
        "header is different annotations. e.g. mutation status.",
    )
    outliers.add_argument(
        "--output_prefix",
        type=_check_output_prefix,
        default="outliers",
        help="Output prefix for writing files. Default outliers. ",
    )
    outliers.add_argument(
        "--iqrs",
        type=_check_positive,
        default=1.5,
        help="Number of inter-quartile ranges (IQRs) above or below the "
        "median to consider a value an outlier. Default is 1.5.",
    )
    outliers.add_argument(
        "--up_or_down",
        type=str,
        default="true",
        choices=["up", "down"],
        help="Whether to look for up or down outliers. Choices are up or down. Default up.",
    )
    outliers.add_argument(
        "--do_not_aggregate",
        default=False,
        action="store_true",
        help="Use flag if you do not want to sum outliers based on site prefixes.",
    )
    outliers.add_argument(
        "--write_outlier_table",
        default=False,
        action="store_true",
        help="Use flag to write a table of outlier counts.",
    )
    outliers.add_argument(
        "--write_frac_table",
        default=False,
        action="store_true",
        help="Use flag if you want to write a table with fraction of "
        "values per site per sample that are outliers. Useful for custom visualization.",
    )
    outliers.add_argument(
        "--ind_sep",
        type=str,
        default="-",
        help="If site labels have a parent molecule (e.g. a gene name such "
        "as ATM) and a site identifier (e.g. S365) this is the "
        "delimiter between the two elements. Default is -",
    )
    outliers.add_argument(
        "--frac_filter",
        type=_bn0and1,
        default=0.3,
        help="The minimum fraction of samples per group that must have an outlier in a gene to"
        "consider that gene in the analysis. This is used to prevent a high number of outlier "
        "values in 1 sample from driving a low qvalue. Default 0.3",
    )
    outliers.add_argument(
        "--write_comparison_summaries",
        default=False,
        action="store_true",
        help="Use flag to write a separate file for each column in the annotations table, "
        "with outlier counts in each group, p-values and q-values in each group. ",
    )
    outliers.add_argument(
        "--fdr",
        type=_bn0and1,
        default=0.05,
        help="FDR threshold to use to select genes to visualize. Default 0.05",
    )
    outliers.add_argument(
        "--write_gene_list",
        default=False,
        action="store_true",
        help="Use flag to write a list of significantly enriched genes for each value in each "
        "comparison.",
    )
    outliers.add_argument(
        "--make_heatmaps",
        default=False,
        action="store_true",
        help="Use flag to draw a heatmap of significantly enriched genes for each value in each "
        "comparison. If used, need an fdr threshold as well. ",
    )
    outliers.add_argument(
        "--red_or_blue",
        type=str,
        choices=["red", "blue"],
        default="red",
        help="Color of values to draw on heatmap. Default red. ",
    )
    outliers.add_argument(
        "--annotation_colors",
        type=_is_valid_file,
        default=None,
        help="File with color map to use for annotation header. Must "
        "have a line with 'value    color' format for each value in annotations. Any value "
        "not represented will be assigned a new color. ",
    )
    return parser


# Module runner
def _main(args: Optional[List[str]] = None):
    if args is None:
        args = sys.argv[1:]
    args = _make_parser().parse_args(args)

    logger = _set_up_logger(args.output_prefix)

    logger.info("Running deva in %s mode" % args.which)
    for arg in vars(args):
        logger.info("Parameter %s: %s" % (arg, getattr(args, arg)))

    if args.which == "outliers_table":
        df = parsers.read_in_values(args.values)
        make_outliers_table(
            df,
            iqrs=args.iqrs,
            up_or_down=args.up_or_down,
            aggregate=~args.do_not_aggregate,
            save_outlier_table=True,
            save_frac_table=args.write_frac_table,
            output_prefix=args.output_prefix,
            ind_sep=args.ind_sep,
        )

    elif args.which == "binarize":
        annotations = parsers.read_in_values(args.annotations)
        annotations = parsers.binarize_annotations(annotations)
        annotations.to_csv("%s.binarized.tsv" % args.output_prefix, sep="\t")

    elif args.which == "normalize":
        target = parsers.read_in_values(args.target_values)
        normalizer = parsers.read_in_values(args.normalizer_values)
        df = parsers.normalize_df(target, normalizer, args.ind_sep)
        df.to_csv(args.output_prefix + ".normalized.tsv", sep='\t')

    elif args.which == "compare_groups":
        outliers = parsers.read_in_outliers(
            args.outliers_table, args.up_or_down, args.iqrs
        )

        annotations = parsers.read_in_values(args.annotations)
        qVals = compare_groups_outliers(
            outliers,
            annotations,
            frac_filter=args.frac_filter,
            save_qvalues=True,
            output_prefix=args.output_prefix,
            save_comparison_summaries=args.write_comparison_summaries,
        )
        if args.write_gene_list:
            qVals.write_gene_lists(args.fdr, args.output_prefix)

        if args.make_heatmaps:
            for col_of_interest in qVals.df.columns:
                plot_heatmap(
                    annotations,
                    qVals.df,
                    col_of_interest,
                    outliers.frac_table,
                    fdr=args.fdr,
                    red_or_blue=args.red_or_blue,
                    output_prefix=args.output_prefix,
                    colors=args.annotation_colors,
                    savefig=True,
                )
                plt.close()

    elif args.which == "visualize":
        qvals = parsers.read_in_values(args.comparison_qvalues)
        annotations = parsers.read_in_values(args.annotations)
        frac_table = parsers.read_in_values(args.visualization_table)
        col_of_interest = args.comparison_of_interest
        annot_cols = args.annotations_to_show[0].split()

        try:
            annotations = annotations[annot_cols]
        except KeyError:
            raise ValueError("Some of %s not in annotations" % annot_cols)

        annot_label = col_of_interest.split("_", 1)[1].rsplit("_", 1)[0]
        if annot_label not in annot_cols:
            raise ValueError("%s must be in annotations columns to show" % annot_label)

        plot_heatmap(
            annotations,
            qvals,
            col_of_interest,
            frac_table,
            fdr=args.fdr,
            red_or_blue=args.red_or_blue,
            output_prefix=args.output_prefix,
            colors=args.annotation_colors,
            savefig=True,
        )
        plt.close()
        if args.write_gene_list:
            qValues(qvals[[col_of_interest]], [], None).write_gene_lists(
                args.fdr, args.output_prefix
            )

    elif args.which == "deva":
        df = parsers.read_in_values(args.values)
        annotations = parsers.read_in_values(args.annotations)
        outLiers, qVals = deva(
            df,
            annotations,
            iqrs=args.iqrs,
            frac_filter=args.frac_filter,
            up_or_down=args.up_or_down,
            aggregate=~args.do_not_aggregate,
            save_outlier_table=args.write_outlier_table,
            save_frac_table=args.write_frac_table,
            save_qvalues=True,
            output_prefix=args.output_prefix,
            ind_sep=args.ind_sep,
            save_comparison_summaries=args.write_comparison_summaries,
        )
        if args.write_gene_list:
            qVals.write_gene_lists(args.fdr, args.output_prefix)

        if args.make_heatmaps:
            for col_of_interest in qVals.df.columns:
                plot_heatmap(
                    annotations,
                    qVals.df,
                    col_of_interest,
                    outLiers.frac_table,
                    fdr=args.fdr,
                    red_or_blue=args.red_or_blue,
                    output_prefix=args.output_prefix,
                    colors=args.annotation_colors,
                    savefig=True,
                )
                plt.close()

    with open(parameters_file_name % args.output_prefix, "w") as fh:
        for arg in vars(args):
            fh.write("%s: %s\n" % (arg, getattr(args, arg)))


# Run cli
if __name__ == "__main__":
    _main()



# Used in parsers
binarized_col_name = "%s_%s"  # % (col, val)
outgroup_val = "not-%s"  # % val


# Used primarily in outlierTable
row_iqr_name = "row_iqr"
row_median_name = "row_median"
row_upper_bound_name = "row_medPlus"
row_lower_bound_name = "row_medMinus"
col_seps = "_"
col_not_outlier_suffix = "notOutliers"
col_outlier_suffix = "outliers"
agg_col = "gene"


# Used primarily in comparisons
outlier_count_lab = "Outliers"
not_outlier_count_lab = "NotOutlier"
fisherp_col = "fisherp"
fisherfdr_col = "fisherFDR"
mult_hypoth_method = "fdr_bh"


# Used in outliers
fdr_col_label = "fisherFDR_%s_%s"  # % (comp, group_label)
general_group_label_0 = "0"
general_group_label_1 = "1"
comp_group_suffix = "_%s_%s"
general_fisher_p = "fisherp"
specific_fisher_p = "fisherp_%s_%s"


# Used in visualization
default_palette = "Set2"
cbar_label = "Fraction\nOutliers"
plot_title = "Outliers in %s"


# File name outputs
gene_list_file_name = "%s.%s.sig_genes.fdr%s.txt"
frac_table_file_name = "%s.%s.fraction_table.tsv"
outlier_table_file_name = "%s.%s.count_table.tsv"
ind_comparison_file_name = "%s.%s.%s.qvalues.tsv"
qvalues_file_name = "%s.%s.qvalues.tsv"
figure_file_name = "%s.%s.fdr%s.heatmap.pdf"
parameters_file_name = "%s.parameters.txt"



from typing import List, Optional, Iterable
import logging
import pandas as pd
from pandas import DataFrame
import numpy as np
from blacksheep._constants import col_seps, col_outlier_suffix, col_not_outlier_suffix, gene_list_file_name


def list_to_file(lis: Iterable, filename: str):
    """Takes an iterable and a file path and writes a value per line from the iterable into the new
    file.

    Args:
        lis: Iterable to write to file
        filename: Filename to write to.

    Returns:
        None

    """

    with open(filename, "w") as fh:
        for x in lis:
            fh.write("%s\n" % x)


def make_frac_table(df, samples):
    """Constructs the fraction table from the outliers table

    Returns: A DataFrame with one column per sample, with the fraction of outliers per row
    per sample. This table is useful for visualization but not statistics.

    """
    df = df.copy()
    cols_outliers = [x + col_seps + col_outlier_suffix for x in samples]
    cols_notOutliers = [x + col_seps + col_not_outlier_suffix for x in samples]
    df = df.fillna(0)
    num_total_psites = df[cols_notOutliers].values + df[cols_outliers].values
    with np.errstate(divide="ignore", invalid="ignore"):
        frac_table = df[cols_outliers].values / num_total_psites

    return DataFrame(frac_table, index=df.index, columns=samples)

class OutlierTable:
    """Output of calling outliers. """

    def __init__(
        self,
        df: DataFrame,
        updown: str,
        iqrs: Optional[float],
        samples: Optional[list],
        frac_table: Optional[DataFrame],
    ):
        """Instantiate an OutlierTable

        Args:
            df: DataFrame with outlier and non-outlier columns, and genes/sites as rows.
            updown: Whether the outliers are above or below the median. Options are "up" or "down"
            iqrs: The IQR threshold used to call outliers.
            samples: The samples included in the analysis to define median and IQR.
            frac_table: DataFrame with samples as columns and genes/sites as rows indicating
            what fraction of sites per sample were called as outliers. Useful for visualization.
        """

        self.df = df
        self.up_or_down = updown
        self.iqrs = iqrs
        self.samples = samples
        if frac_table is not None:
            self.frac_table = frac_table
        else:
            self.frac_table = make_frac_table(df, samples)



class qValues:
    """Output from comparing groups using outliers. """

    def __init__(self, df: DataFrame, comps: list, frac_filter: Optional[float]):
        """Instantiates a qValues object.

        Args:
            df: DataFrame with genes/sites as rows and comparison_group as columns.
            comps: List of comparisons used to populate table.
            frac_filter: What fraction of samples in group of interest were required to have
            an outliers for any given row to be considered for analysis.
        """

        self.df = df
        self.comps = comps
        self.frac_filter = frac_filter

    def write_gene_lists(
        self,
        fdr_cut_off: float = 0.01,
        output_prefix: str = "outliers",
        comparisons: Optional[List] = None,
    ):
        """ Writes significant gene list files for every column in a qvalue table

        Args:
            fdr_cut_off: FDR threshold for significance
            output_prefix: Output prefix for files
            comparisons: which subset of qvalue columns to write gene lists for. Default will
            write for all columns

        Returns: None

        """

        if comparisons is None:
            comparisons = self.df.columns
        else:
            comparisons = [
                col
                for col in self.df.columns
                if (col.startswith(tuple(comparisons))) and (col in self.comps)
            ]

        for comp in comparisons:
            sig_genes = list(self.df.loc[(self.df[comp] < fdr_cut_off), :].index)
            list_to_file(
                sig_genes, gene_list_file_name % (output_prefix, comp, fdr_cut_off)
            )


    def make_signed_logqs(self) -> DataFrame:
        """Create a DataFrame with signed log10 qvalues for each comparison. E.g. group1 qvalues
        will be positive, and group 2 qvalues will be negative. Assignment of positive group is
        based on order in qvalues, could be helpful to negate some columns in output depending on
        group of interest.

        Returns: DataFrame with signed qvalues.

        """
        if not (self.comps is None):
            self.comps = [i.split('_', 1)[1].rsplit('_', 1)[0] for i in self.df.columns]
            self.comps = sorted(list(set(self.comps)))

        signed_qs = pd.DataFrame()
        for comp in self.comps:
            cols = [
                col for col in self.df.columns if col.split('_', 1)[1].rsplit('_', 1)[0] == comp
            ]

            if len(cols) > 2 or len(cols) == 0:
                logging.warning("Excluding %s, %s columns are associated with %s, need 1 or 2 "
                                "columns. Annotation value probably has an _ in it. "
                                %(comp, len(cols), comp))
                continue

            if len(cols) == 1:
                temp = self.df[cols[0]]
            elif len(cols) == 2:
                temp = pd.DataFrame(
                    -np.log10(self.df[cols[1]]).subtract(np.log10(self.df[cols[0]]), fill_value=0),
                    columns=['%s_%s'%(comp, cols[1].rsplit('_', 1)[1])]
                )
            signed_qs = pd.concat([signed_qs, temp], join='outer', axis=1, sort=False)

        return signed_qs



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="blksheep",
    version="0.0.3",
    author="Ruggles Lab",
    author_email="ruggleslab@gmail.com",
    description="A package for differential extreme values analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ruggleslab/blackSheep/",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    install_requires=[
        "pandas >= 0.24.2",
        "numpy >= 1.16.4",
        "scipy >= 1.2.1",
        "matplotlib >= 3.1.0",
        "seaborn >= 0.9.0",
        "scikit-learn >= 0.21.2",
        "statsmodels >= 0.10.0"
    ],
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    entry_points={"console_scripts": ["blacksheep = blacksheep.cli:_main"]},
)



import logging
from typing import List, Tuple, Iterable, Optional
import pandas as pd
from pandas import DataFrame
from pandas import Series
import scipy.stats
from statsmodels.stats.multitest import multipletests
from blacksheep._constants import *


SampleList = List[str]
logger = logging.getLogger("cli")


def get_sample_lists(
    annotations: DataFrame, col: str
) -> Tuple[Optional[str], Optional[SampleList], Optional[str], Optional[SampleList]]:
    """Finds groupings of samples from an annotation DataFrame column.

    Args:
        annotations: A DataFrame with samples as the index and annotations as columns. Each
        column must contain exactly 2 different values, and optionally missing values. Columns
        with less or more than 2 options will be ignored.
        col: Which column for which to define groups.

    Returns: A label for group0, the list of samples in group0, a label for group1 and the list
        of samples in group1.

    """

    groups = list(pd.Series(annotations[col].value_counts().keys()).dropna())
    if len(groups) != 2:
        return None, None, None, None
    group0 = list(annotations.loc[annotations[col] == groups[0], :].index)
    group1 = list(annotations.loc[annotations[col] == groups[1], :].index)
    return groups[0], group0, groups[1], group1


def _filter_outliers(
    df: DataFrame,
    group0_list: SampleList,
    group1_list: SampleList,
    frac_filter: Optional[float],
) -> DataFrame:
    """Filters an outlier count table for rows that are enriched for outliers in group0 and that
    have more than a frac_filter fraction of samples of group0 with an outlier.

    Args:
        df: Outliers count table, output from convertToCounts. Samples are columns,
        genes/sites are the index.
        group0_list: List of samples in the group of interest.
        group1_list: List of samples in the outgroup.
        frac_filter: The fraction of samples in group0 (i.e. the group of interest) that must
        have an outlier value to be considered in the comparison. Float between 0 and 1 or None.

    Returns: A DataFrame with rows that are not enriched in group0 removed. If frac_filter > 0,
    rows without enough outliers in group0 are also removed.

    """
    df = df.copy()
    group0_outliers = [x + col_seps + col_outlier_suffix for x in group0_list]
    group0_notOutliers = [x + col_seps + col_not_outlier_suffix for x in group0_list]
    group1_outliers = [x + col_seps + col_outlier_suffix for x in group1_list]
    group1_notOutliers = [x + col_seps + col_not_outlier_suffix for x in group1_list]

    if (frac_filter < 0) or (frac_filter > 1):
        raise ValueError("Frac filter must be between 0 and 1")
    if frac_filter != None:
        min_num_outlier_samps = len(group0_list) * frac_filter
        num_outlier_samps = (df[group0_outliers] > 0).sum(axis=1)
        df = df.loc[num_outlier_samps >= min_num_outlier_samps, :]

    # Filter for higher proportion of outliers in group0 than group1
    group0_outlier_rate = (
        df[group0_outliers]
        .sum(axis=1)
        .divide(df[group0_outliers + group0_notOutliers].sum(axis=1), axis=0)
    )
    group1_outlier_rate = (
        df[group1_outliers]
        .sum(axis=1)
        .divide(df[group1_outliers + group1_notOutliers].sum(axis=1), axis=0)
    )

    df = df.loc[group0_outlier_rate > group1_outlier_rate, :]

    return df


def _fisher_test_groups(
    group0_list: SampleList,
    group1_list: SampleList,
    outlier_table: DataFrame,
    correction_type: str = mult_hypoth_method,
) -> Tuple[Series, DataFrame]:
    """Performs fishers test by counting outlier and not outlier sites in two groups. Corrects for
    multiple hypothesis testing.

    Args:
        group0_list: List of samples in group of interest
        group1_list: List of samples in outgroup
        outlier_table: Outlier count table, like output of convertToCounts
        correction_type: Method to use for multiple hypothesis correction.

    Returns: Series of qvalues with index matching filtered rows.

    """

    outliers_group0_list = [x + col_seps + col_outlier_suffix for x in group0_list]
    notOutliers_group0_list = [
        x + col_seps + col_not_outlier_suffix for x in group0_list
    ]
    outliers_group1_list = [x + col_seps + col_outlier_suffix for x in group1_list]
    notOutliers_group1_list = [
        x + col_seps + col_not_outlier_suffix for x in group1_list
    ]

    outlier_table[outlier_count_lab + general_group_label_0] = outlier_table[
        outliers_group0_list
    ].sum(axis=1)
    outlier_table[not_outlier_count_lab + general_group_label_0] = outlier_table[
        notOutliers_group0_list
    ].sum(axis=1)
    outlier_table[outlier_count_lab + general_group_label_1] = outlier_table[
        outliers_group1_list
    ].sum(axis=1)
    outlier_table[not_outlier_count_lab + general_group_label_1] = outlier_table[
        notOutliers_group1_list
    ].sum(axis=1)

    outlier_table[fisherp_col] = outlier_table.apply(
        (
            lambda r: scipy.stats.fisher_exact(
                [
                    [
                        r[outlier_count_lab + general_group_label_0],
                        r[outlier_count_lab + general_group_label_1],
                    ],
                    [
                        r[not_outlier_count_lab + general_group_label_0],
                        r[not_outlier_count_lab + general_group_label_1],
                    ],
                ]
            )[1]
        ),
        axis=1,
    )

    outlier_table[fisherfdr_col] = multipletests(
        list(outlier_table[fisherp_col]), method=correction_type
    )[1]
    fisher_info = outlier_table[
        [
            outlier_count_lab + general_group_label_0,
            outlier_count_lab + general_group_label_1,
            not_outlier_count_lab + general_group_label_0,
            not_outlier_count_lab + general_group_label_1,
            fisherp_col,
        ]
    ]
    return outlier_table[fisherfdr_col], fisher_info


def _compare_groups(
    results_df: DataFrame,
    outliers: DataFrame,
    group0: SampleList,
    group1: SampleList,
    frac_filter: Optional[float],
    label: str,
) -> Tuple[DataFrame, Optional[DataFrame]]:
    """Performs fisher test and cleans up a fisher infor table for making output for each comparison

    Args:
        results_df: Accumulating qvalues DataFrame
        outliers: Outliers count DataFrame
        group0: List of samples in group of interest
        group1: List of samples in outgroup
        frac_filter: Fraction of samples in group of interest require to have an outlier per
    site to be considered in analysis
        label: What to call the FDR output column on the qvalues DataFrame

    Returns: Concatenated qvalues DataFrame and a table of info about the comparison

    """

    df = _filter_outliers(outliers, group0, group1, frac_filter)
    logger.info("Calculating enrichment in %s rows for %s" % (len(df), label))
    if len(df) > 0:
        col, fisher_info = _fisher_test_groups(group0, group1, df)
        col = DataFrame(col)
        col.columns = [label]
        results_df = pd.concat([results_df, col], axis=1, join="outer", sort=False)
    else:
        logger.warning("No rows tested for %s" % label)
        fisher_info = DataFrame(
            columns=[
                outlier_count_lab + general_group_label_0,
                outlier_count_lab + general_group_label_1,
                not_outlier_count_lab + general_group_label_0,
                not_outlier_count_lab + general_group_label_1,
                fisherp_col,
            ]
        )
    return results_df, fisher_info
